package cn.sagacloud.server.algorithm.common.utils;

import cn.sagacloud.server.algorithm.models.dataCheck.LevelCheckResult;
import cn.sagacloud.server.algorithm.models.modelFile.FloorModelView;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeanUtils;

import java.util.*;

/**
 * 楼层号检查
 * @author Administrator
 * @date 2021-10-21
 */
public class FloorLevelCheckUtil {

    /**
     * 检查
     * @param views
     * @return
     */
    public static List<LevelCheckResult> checkLevel(List<FloorModelView> views) {
        if(CollectionUtils.isEmpty(views)) {
            return null;
        }
        // 排序
        FloorModelView[] viewArr = views.toArray(new FloorModelView[0]);
        Arrays.sort(viewArr, new FloorModelCompare());
        Set<String> errorIds = new HashSet<>();
        FloorModelView lastView = null;
        // 检查有楼层号异常的楼层
        for(FloorModelView view:viewArr) {
            Double level = view.getBaseLevel();
            if(level != null && (lastView == null || level > lastView.getBaseLevel())) {
                lastView = view;
                continue;
            }
            // 将前后的模型ID放入异常队列
            errorIds.add(lastView.getCurrentModelId());
            errorIds.add(view.getCurrentModelId());
            lastView = null;
        }
        // 回写结果，按原顺序
        List<LevelCheckResult> result = new ArrayList<>();
        for(FloorModelView view:views) {
            LevelCheckResult rs = new LevelCheckResult();
            BeanUtils.copyProperties(view, rs);
            rs.setHasError(errorIds.contains(view.getCurrentModelId()));
            rs.setErrorMsg(rs.getHasError()?"标高冲突":null);
            result.add(rs);
        }
        return result;
    }

    /**
     * 楼层比较器
     */
    protected static class FloorModelCompare implements Comparator<FloorModelView> {

        @Override
        public int compare(FloorModelView v1, FloorModelView v2) {
            if(v1 == null || v2 == null) {
                return v1 == null ? 1 : -1;
            }
            FloorLevel o1 = convertToFloorLevel(v1.getFloorName());
            FloorLevel o2 = convertToFloorLevel(v2.getFloorName());
            // 如果楼层类型不一样
            if(o1.type != o2.type) {
                return o1.type - o2.type;
            }
            // 楼层类型一样，则看层数
            if(o1.level != o2.level) {
                // 层号越大的越高，但地下则刚好相反
                int result = o1.level - o2.level;
                return o1.type == 0 ? -result:result;
            }
            // 都一样则看夹层号，夹层号越大越高
            if(o1.middle != o2.middle) {
                return o1.middle - o2.middle;
            }
            return 0;
        }


        /**
         * 根据楼层名称转换楼层信息
         * @param floorName
         * @return
         */
        protected FloorLevel convertToFloorLevel(String floorName) {
            FloorLevel floor = new FloorLevel();
            if(StringUtils.isBlank(floorName)) {
                return floor;
            }
            // 顶层夹层
            if(floorName.startsWith("RFM")) {
                floor.type = 2;
                if(floorName.length() > 3) {
                    floor.middle = Integer.valueOf(floorName.substring(3));
                }
                return floor;
            }
            // 顶层
            if(floorName.startsWith("RF")) {
                floor.type = 2;
                return floor;
            }
            // 正常楼层/地下
            if(floorName.startsWith("F") || floorName.startsWith("B")) {
                floor.type = floorName.startsWith("F")?1:0;
                // 是否有夹层
                int mIndex = floorName.indexOf("M");
                if(mIndex < 0) {
                    floor.level = Integer.valueOf(floorName.substring(1));
                } else if (floorName.endsWith("M")) {
                    // 没夹层数也算没夹层
                    floor.level = Integer.valueOf(floorName.substring(1, mIndex));
                } else {
                    floor.level = Integer.valueOf(floorName.substring(1, mIndex));
                    floor.middle = Integer.valueOf(floorName.substring(mIndex+1));
                }
                return floor;
            }
            return floor;
        }
    }

    /**
     * 楼层信息
     */
    public static class FloorLevel {
        /** 楼层类型：0-地下，1-正常，2-顶层 */
        public int type = 1;
        /** 楼层号 */
        public int level = 1;
        /** 夹层号 */
        public int middle = 0;
    }
}
