package cn.sagacloud.server.algorithm.common.utils

import cn.sagacloud.server.algorithm.backstage.model.datacenter.DCFloor
import cn.sagacloud.server.algorithm.backstage.model.datacenter.DCISpace
import cn.sagacloud.server.algorithm.backstage.model.datacenter.QueryResponse
import cn.sagacloud.server.algorithm.backstage.model.datacenter.UpdateRequest
import cn.sagacloud.server.algorithm.models.entities.Space
import cn.sagacloud.server.algorithm.models.jsonAnalyzer.Point
import cn.sagacloud.server.algorithm.models.modelFile.ModelFile
import cn.sagacloud.server.algorithm.services.BaseDataService.modelFileService
import cn.sagacloud.server.algorithm.services.BaseDataService.spaceService
import cn.sagacloud.server.algorithm.services.datacenter.DataCenterRequest
import com.sybotan.service.SObjectService
import com.sybotan.service.models.SCascadeQuery
import com.sybotan.service.models.requests.SQueryRequest
import com.sybotan.service.models.responses.SQueryResponse
import org.apache.commons.collections4.CollectionUtils
import org.apache.commons.lang3.StringUtils
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.stereotype.Component
import java.awt.Polygon
import java.util.*
import java.util.stream.Collectors

@Component
class SpaceAffectUtil {

    @Autowired
    private val service: DataCenterRequest? = null
    fun affectedSpace(projectId: String, modelId: String): Boolean {
        if (StringUtils.isBlank(projectId) || StringUtils.isBlank(modelId)) {
            return true
        }
        // 查楼层
        val spaceList = querySpaceByFloorModel(projectId, modelId)
        if (CollectionUtils.isEmpty(spaceList)) {
            return true
        }
        // 查原模型
        val modelIds = queryModelFile(projectId, modelId)
        if (CollectionUtils.size(modelIds) != 2) {
            return true
        }
        // 查模型关联的元空间
        val iSpaceMap = queryMetaSpaceByModel(modelIds)
        if (CollectionUtils.isEmpty(iSpaceMap[modelIds!![1]])) {
            // 旧模型没有关联的元空间不需要处理
            return true
        }
        // 获取有变化的空间列表
        val spaces = getAffectedSpaceIds(spaceList, iSpaceMap[modelIds[0]], iSpaceMap[modelIds[1]])
        // 更新空间
        updateSpaceList(projectId, spaces)
        return true
    }

    /**
     * 获取有变化的空间ID列表
     * @param spaces
     * @param newMetaSpaces
     * @param oldMetaSpaces
     * @return
     */
    private fun getAffectedSpaceIds(spaces: List<DCISpace?>?, newMetaSpaces: List<Space>?, oldMetaSpaces: List<Space>?): List<DCISpace?>? {
        // 准备修改后的元空间
        val newMetaSpaceMap: MutableMap<String?, Space> = HashMap()
        var i = 0
        val j = CollectionUtils.size(newMetaSpaces)
        while (i < j) {
            val space = newMetaSpaces!![i]
            newMetaSpaceMap[space.id] = space
            i++
        }
        // 截取有变化的元空间
        val affectList: MutableList<Polygon?> = ArrayList()
        for (space in oldMetaSpaces!!) {
            val polygon = createPolygon(space.outline)
            val newSpace = newMetaSpaceMap[space.id]
            if (newSpace == null) {
                affectList.add(polygon)
                continue
            }
            val newPolygon = createPolygon(newSpace.outline)
            // TODO 有交集则加入有变化的元空间列表
            if (newPolygon == null) {
                affectList.add(polygon)
                continue
            }
        }
        if (CollectionUtils.isEmpty(affectList)) {
            return null
        }
        // 与有变化的元空间有交集的空间
        val rsList: MutableList<DCISpace?> = ArrayList()
        for (space in spaces!!) {
            val polygon = createPolygon(space!!.outline)
            // TODO 有交集则加入有变化的空间列表
            if (polygon == null) {
                // TODO state设置为1
                rsList.add(space)
                continue
            }
        }
        return rsList
    }

    private fun createPolygon(list: ArrayList<ArrayList<Point>>?): Polygon {
        return Polygon()
    }

    /**
     * 更新空间列表
     * @param projectId
     * @param spaces
     */
    private fun updateSpaceList(projectId: String, spaces: List<DCISpace?>?) {
        val request = UpdateRequest<DCISpace>()
        request.content = spaces as ArrayList<DCISpace>
        service!!.iSpaceUpdate(request, projectId)
    }

    /**
     * 查询楼层关联的空间
     * @param projectId
     * @param modelId
     * @return
     */
    private fun querySpaceByFloorModel(projectId: String, modelId: String): List<DCISpace?>? {
        val request = SQueryRequest()
        request.pageSize = 500
        request.filters = " model_id = '$modelId' "
        val cascade = SCascadeQuery()
        cascade.name = "spaceList"
        val cascades = ArrayList<SCascadeQuery>()
        cascades.add(cascade)
        request.cascade = cascades
        val response: QueryResponse<DCFloor>? = service!!.floorQuery(request, projectId)
        if (response == null || CollectionUtils.isEmpty(response.content)) {
            return null
        }
        val floors = response.content
        val spaceList: MutableList<DCISpace> = ArrayList<DCISpace>()
        for (floor in floors!!) {
            val spaces: List<DCISpace>? = floor!!.spaceList
            if (CollectionUtils.isEmpty(spaces)) {
                continue
            }
            spaceList.addAll(spaces!!)
        }
        return spaceList
    }

    /**
     * 获取模型
     * @param projectId
     * @param modelId
     * @return
     */
    private fun queryModelFile(projectId: String, modelId: String): List<String>? {
        val modelFileService: SObjectService<ModelFile> = modelFileService
        val request = SQueryRequest()
        request.filters = "model_floor_id = '$modelId' and removed = false and status = 4 and version is not null"
        request.orders = "version desc"
        val projection = ArrayList<String>()
        projection.add("id")
        request.projection = projection
        val response: SQueryResponse<ModelFile> = modelFileService.pageQuery(request, null)
        if (response == null || CollectionUtils.isEmpty(response.content)) {
            return null
        }
        val floors = response.content
        return floors!!.stream().map<String>(ModelFile::id).collect(Collectors.toList())
    }

    /**
     * 根据模型查询关联的元空间
     * @param modelIds
     * @return
     */
    private fun queryMetaSpaceByModel(modelIds: List<String>?): Map<String, List<Space>?> {
        val spaceMap: MutableMap<String, List<Space>?> = HashMap(2)
        val spaceService: SObjectService<Space> = spaceService
        // 查询新模型的元空间
        var request = SQueryRequest()
        request.filters = "model_id = '" + modelIds!![0] + "' "
        val projection = ArrayList<String>()
        projection.add("sourceId")
        projection.add("outline")
        request.projection = projection
        var response: SQueryResponse<Space> = spaceService.pageQuery(request, null)
        if (CollectionUtils.isNotEmpty(response.content)) {
            spaceMap[modelIds[0]] = response.content
        }
        // 查询旧模型的元空间
        request = SQueryRequest()
        request.filters = "model_id = '" + modelIds[1] + "' "
        response = spaceService.pageQuery(request, null)
        if (CollectionUtils.isNotEmpty(response.content)) {
            spaceMap[modelIds[1]] = response.content
        }
        return spaceMap
    }
}